import json
import os

import pandas as pd
from markupsafe import escape
from featureform import ResourceClient
from featureform.serving import LocalClientImpl
from flask import Blueprint, Response, request
from flask_cors import CORS, cross_origin
from .metadata_repository import (
    MetadataRepositoryLocalImpl,
    Feature,
    FeatureVariant,
    TrainingSet,
    TrainingSetVariant,
    Source,
    SourceVariant,
    Label,
    LabelVariant,
    Entity,
    Model,
    User,
    Provider,
)
from .sqlite_metadata import SQLiteMetadata
from .type_objects import (
    FeatureResource,
    FeatureVariantResource,
    TrainingSetResource,
    TrainingSetVariantResource,
    SourceResource,
    SourceVariantResource,
    EntityResource,
    UserResource,
    ModelResource,
    LabelResource,
    LabelVariantResource,
    ProviderResource,
)
from .resources import SourceType
from .version import get_package_version

path = os.path.join(os.path.dirname(__file__), "dashboard")

dashboard_app = Blueprint(
    "dashboard_app", __name__, static_folder=path + "/out/", static_url_path=""
)

CORS(dashboard_app)


@dashboard_app.route("/")
def index():
    return dashboard_app.send_static_file("index.html")


@dashboard_app.route("/<type>")
def get_type(type):
    return dashboard_app.send_static_file("[type].html")


@dashboard_app.route("/<type>/<entity>")
def entity(type, entity):
    return dashboard_app.send_static_file("[type]/[entity].html")


@dashboard_app.route("/search")
def search():
    return dashboard_app.send_static_file("query.html")


@dashboard_app.route("/query")
def query():
    return dashboard_app.send_static_file("query.html")


@dashboard_app.route("/static/<asset>")
def deliver_static(asset):
    return dashboard_app.send_static_file("static/" + asset)


@dashboard_app.route("/data/version")
def version():
    return {"version": get_package_version()}


@dashboard_app.route("/data/sourcedata", methods=["GET"])
@cross_origin(allow_headers=["Content-Type"])
def source_data():
    limit = 150
    n = 0
    name = request.args["name"]
    variant = request.args["variant"]
    if name == "" or variant == "":
        error = f"Error 400: GetSourceData - Could not find the name({name}) or variant({variant}) query parameters."
        return Response(
            response=json.dumps(error), status=400, mimetype="application/json"
        )
    try:
        with LocalClientImpl() as localClientImpl:
            source_data = {"columns": [], "rows": []}
            df = localClientImpl.get_input_df(name, variant)
            if isinstance(df, pd.Series):
                df = df.to_frame()
                df.reset_index(inplace=True)
            for column in df.columns:
                source_data["columns"].append(column)
            for _, currentRow in df.iterrows():
                n = n + 1
                if n > limit:
                    break
                currentRow = currentRow.fillna("NaN")
                source_data["rows"].append(currentRow.to_list())
            return json.dumps(source_data, allow_nan=False)
    except Exception as e:
        print(e)
        error = "Error 500: Unable to retrieve source_data columns."
        return Response(
            response=json.dumps(error), status=500, mimetype="application/json"
        )


@dashboard_app.route("/data/<type>/<resource>/gettags", methods=["POST"])
@cross_origin(allow_headers=["Content-Type"])
def get_tags(type, resource):
    try:
        response = {"name": resource, "variant": request.json["variant"], "tags": []}
        db = MetadataRepositoryLocalImpl(SQLiteMetadata())
        tags = db.get_tags_for_resource(
            escape(response["name"]), escape(response["variant"]), type
        )
        if len(tags):
            response["tags"] = json.loads(tags)
        return json.dumps(response, allow_nan=False)
    except Exception as e:
        print(e)
        error = f"Error 500: Unable to retrieve tags from resource."
        return Response(
            response=json.dumps(error), status=500, mimetype="application/json"
        )


@dashboard_app.route("/data/<type>/<resource>/tags", methods=["POST"])
@cross_origin(allow_headers=["Content-Type"])
def post_tags(type, resource):
    try:
        response = {
            "name": resource,
            "variant": request.json["variant"],
            "tags": request.json["tags"],
        }
        db = MetadataRepositoryLocalImpl(SQLiteMetadata())

        if type == "features":
            found_resource = db.get_feature_variant(
                response["name"], response["variant"]
            )
        elif type == "training_sets":
            found_resource = db.get_training_set_variant(
                response["name"], response["variant"]
            )
        elif type == "sources":
            found_resource = db.get_source_variant(
                response["name"], response["variant"]
            )
        elif type == "labels":
            found_resource = db.get_label_variant(response["name"], response["variant"])
        elif type == "entities":
            found_resource = db.get_entity(response["name"])
        elif type == "models":
            found_resource = db.get_model(response["name"])
        elif type == "users":
            found_resource = db.get_user(response["name"])
        elif type == "providers":
            found_resource = db.get_provider(response["name"])

            found_resource.tags = escape(response["tags"])
            db.update_resource(found_resource)
        return json.dumps(response, allow_nan=False)
    except Exception as e:
        print(e)
        error = f"Error 500: Unable to post tags to resource."
        return Response(
            response=json.dumps(error), status=500, mimetype="application/json"
        )


@dashboard_app.route("/data/<type>", methods=["POST", "GET"])
@cross_origin(allow_headers=["Content-Type"])
def get_metadata_list(type):
    type = type.replace("-", "_")
    if type not in [
        "features",
        "training_sets",
        "sources",
        "labels",
        "entities",
        "models",
        "users",
        "providers",
    ]:
        error_data = f"Invalid resource type: {type}"
        response = Response(
            response=json.dumps(error_data), status=400, mimetype="application/json"
        )
        return response

    db = MetadataRepositoryLocalImpl(SQLiteMetadata())
    metadata_object_list = []
    if type == "features":
        for feature in db.get_features():
            metadata_object_list.append(build_feature_resource(feature))
    elif type == "training_sets":
        for training_set in db.get_training_sets():
            metadata_object_list.append(build_training_set_resource(training_set))
    elif type == "sources":
        for source in db.get_sources():
            metadata_object_list.append(build_source_resource(source))
    elif type == "labels":
        for label in db.get_labels():
            metadata_object_list.append(build_label_resource(label))
    elif type == "entities":
        for entity in db.get_entities():
            metadata_object_list.append(build_entity_resource(entity))
    elif type == "models":
        for model in db.get_models():
            metadata_object_list.append(build_model_resource(model))
    elif type == "users":
        for user in db.get_users():
            metadata_object_list.append(build_user_resource(user))
    elif type == "providers":
        for provider in db.get_providers():
            metadata_object_list.append(build_provider_resource(provider))

    response = Response(
        response=json.dumps(metadata_object_list),
        status=200,
        mimetype="application/json",
    )
    return response


@dashboard_app.route("/data/<type>/<resource>", methods=["POST", "GET"])
@cross_origin(allow_headers=["Content-Type"])
def get_metadata(type, resource):
    type = type.replace("-", "_")
    if type not in [
        "features",
        "training_sets",
        "sources",
        "labels",
        "entities",
        "models",
        "users",
        "providers",
    ]:
        error_data = f"Invalid resource type: {type}"
        response = Response(
            response=json.dumps(error_data), status=400, mimetype="application/json"
        )
        return response

    db = MetadataRepositoryLocalImpl(SQLiteMetadata())
    metadata_object = {}
    if type == "features":
        records = db.get_features()
        records = list(filter(lambda rec: rec.name == resource, records))
        if len(records) > 0:
            metadata_object = build_feature_resource(records[0])
    elif type == "training_sets":
        records = db.get_training_sets()
        records = list(filter(lambda rec: rec.name == resource, records))
        if len(records) > 0:
            metadata_object = build_training_set_resource(records[0])
    elif type == "sources":
        records = db.get_sources()
        records = list(filter(lambda rec: rec.name == resource, records))
        if len(records) > 0:
            metadata_object = build_source_resource(records[0])
    elif type == "labels":
        records = db.get_labels()
        records = list(filter(lambda rec: rec.name == resource, records))
        if len(records) > 0:
            metadata_object = build_label_resource(records[0])
    elif type == "entities":
        records = db.get_entities()
        records = list(filter(lambda rec: rec.name == resource, records))
        if len(records) > 0:
            metadata_object = build_entity_resource(records[0])
    elif type == "models":
        records = db.get_models()
        records = list(filter(lambda rec: rec.name == resource, records))
        if len(records) > 0:
            metadata_object = build_model_resource(records[0])
    elif type == "users":
        records = db.get_users()
        records = list(filter(lambda rec: rec.name == resource, records))
        if len(records) > 0:
            metadata_object = build_user_resource(records[0])
    elif type == "providers":
        records = db.get_providers()
        records = list(filter(lambda rec: rec.name == resource, records))
        if len(records) > 0:
            metadata_object = build_provider_resource(records[0])
    return Response(
        response=json.dumps(metadata_object), status=200, mimetype="application/json"
    )


@dashboard_app.route("/data/search", methods=["GET"])
@cross_origin(allow_headers=["Content-Type"])
def search_metadata():
    raw_query = request.args["q"]
    rc = ResourceClient(local=True)
    results = rc.search(raw_query, True)
    payload = []
    for r in results:
        payload.append(
            {"Name": r["name"], "Variant": r["variant"], "Type": r["resource_type"]}
        )

    return json.dumps(payload)


def variant_list_to_dict(variant_list):
    """Convert a variant list, into a dictionary with variant names as the dictionary keys"""
    variants_dict = dict()

    for variant in variant_list:
        variant_name = variant["variant"]
        variants_dict[variant_name] = variant

    return variants_dict


def resources_list_to_dict(resource_list):
    """Convert a resource list, into a dictionary with resource names as the dictionary keys"""
    variants_dict = dict()

    for variant in resource_list:
        resource_name = variant["name"]
        if resource_name in variants_dict:
            variants_dict[resource_name].append(variant)
        else:
            variants_dict[resource_name] = [variant]

    return variants_dict


def build_feature_resource(feature_main: Feature):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())
    variant_list = []
    for variant_name in feature_main.variants:
        found_variant = db.get_feature_variant(
            name=feature_main.name, variant=variant_name
        )
        variant_list.append(build_feature_variant_resource(found_variant))

    return FeatureResource(
        name=feature_main.name,
        defaultVariant=feature_main.default_variant,
        type="Feature",
        allVariants=feature_main.variants,
        variants=variant_list_to_dict(variant_list),
    ).to_dictionary()


def build_feature_variant_resource(variant_data: FeatureVariant):
    feature_variant_resource = FeatureVariantResource(
        created=variant_data.created,
        description=variant_data.description,
        entity=variant_data.entity,
        name=variant_data.name,
        owner=variant_data.owner,
        provider=variant_data.provider,
        dataType=variant_data.value_type,
        variant=variant_data.variant,
        status=variant_data.status,
        location={
            "entity": variant_data.location.entity,
            "value": variant_data.location.value,
            "timestamp": variant_data.location.timestamp,
        },
        source={
            "Name": variant_data.source[0],
            "Variant": variant_data.source[1],
        },
        tags=variant_data.tags if variant_data.tags is not None else [],
        properties=variant_data.properties
        if variant_data.properties is not None
        else {},
    ).to_dictionary()

    return feature_variant_resource


def build_training_set_resource(training_set_main: TrainingSet):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())
    variant_list = []
    for variant_name in training_set_main.variants:
        found_variant = db.get_training_set_variant(
            name=training_set_main.name, variant=variant_name
        )
        variant_list.append(build_training_set_variant_resource(found_variant))

    return TrainingSetResource(
        name=training_set_main.name,
        defaultVariant=training_set_main.default_variant,
        type="TrainingSet",
        allVariants=training_set_main.variants,
        variants=variant_list_to_dict(variant_list),
    ).to_dictionary()


def build_training_set_variant_resource(variant_data: TrainingSetVariant):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())
    training_set_feature_list = []
    tsf_association_list = db.get_training_set_features(
        name=variant_data.name, variant=variant_data.variant
    )
    for tsf in tsf_association_list:
        found_feature_variant = db.get_feature_variant(
            name=tsf.feature_name, variant=tsf.feature_variant
        )
        training_set_feature_list.append(
            build_feature_variant_resource(found_feature_variant)
        )

    training_set_variant_resource = TrainingSetVariantResource(
        created=variant_data.created if variant_data.created is not None else "",
        description=variant_data.description,
        name=variant_data.name,
        owner=variant_data.owner,
        variant=variant_data.variant,
        label={
            "Name": variant_data.label[0],
            "Variant": variant_data.label[1],
        },
        status=variant_data.status,
        features=resources_list_to_dict(training_set_feature_list),
        tags=variant_data.tags if variant_data.tags is not None else [],
        properties=variant_data.properties
        if variant_data.properties is not None
        else [],
    ).to_dictionary()

    return training_set_variant_resource


def build_source_resource(source_main: Source):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())
    variant_list = []
    for variant_name in source_main.variants:
        found_variant = db.get_source_variant(
            name=source_main.name, variant=variant_name
        )
        variant_list.append(build_source_variant_resource(found_variant))

    return SourceResource(
        name=source_main.name,
        defaultVariant=source_main.default_variant,
        type="Source",
        allVariants=source_main.variants,
        variants=variant_list_to_dict(variant_list),
    ).to_dictionary()


def build_source_variant_resource(variant_data: SourceVariant):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())

    training_set_variant_resource_list = []
    feature_variant_resource_list = []
    try:
        feature_list = db.get_feature_variants_from_source(
            source_name=variant_data.name, source_variant=variant_data.variant
        )
    except ValueError:
        feature_list = []
    for curr_feature in feature_list:
        try:
            # build the feature variant resource
            found_feature_variant = db.get_feature_variant(
                curr_feature.name, curr_feature.variant
            )
            feature_variant_resource_list.append(
                build_feature_variant_resource(found_feature_variant)
            )

            # find the feature training sets, add to trainint set()
            training_set_feature_table_list = db.get_training_set_from_features(
                found_feature_variant.name, found_feature_variant.variant
            )

            for training_set in training_set_feature_table_list:
                found_training_set = db.get_training_set_variant(
                    training_set.training_set_name, training_set.training_set_variant
                )
                training_set_variant_resource_list.append(
                    build_training_set_variant_resource(found_training_set)
                )
        except ValueError:
            continue

    label_variant_resource_list = []
    try:
        label_list = db.get_label_variants_from_source(
            source_name=variant_data.name, source_variant=variant_data.variant
        )
    except:
        label_list = []
    for curr_label in label_list:
        try:
            # build the label variant resource
            label_variant_resource_list.append(build_label_variant_resource(curr_label))

            training_set_variant_list = db.get_training_set_variant_from_label(
                label_name=curr_label.name, label_variant=curr_label.variant
            )

            # only add if not in list. set not hashable
            for tsv in training_set_variant_list:
                if (
                    next(
                        (
                            item
                            for item in training_set_variant_resource_list
                            if (
                                item["name"] == tsv.name
                                and item["variant"] == tsv.variant
                            )
                        ),
                        None,
                    )
                    == None
                ):
                    training_set_variant_resource_list.append(
                        build_training_set_variant_resource(tsv)
                    )
        except ValueError:
            continue

    definition_text = ""
    if variant_data.transformation == SourceType.DF_TRANSFORMATION.value:
        definition_text = db.get_source_variant_text(
            name=variant_data.name, variant=variant_data.variant
        )
    else:
        definition_text = variant_data.definition

    source_variant_resource = SourceVariantResource(
        created=variant_data.created,
        description=variant_data.description,
        name=variant_data.name,
        owner=variant_data.owner,
        definition=definition_text,
        provider=variant_data.provider,
        sourceType=variant_data.source_type,
        variant=variant_data.variant,
        labels=resources_list_to_dict(label_variant_resource_list),
        status=variant_data.status,
        features=resources_list_to_dict(feature_variant_resource_list),
        trainingSets=resources_list_to_dict(training_set_variant_resource_list),
        tags=variant_data.tags if variant_data.tags is not None else [],
        properties=variant_data.properties
        if variant_data.properties is not None
        else [],
    ).to_dictionary()

    return source_variant_resource


def build_label_resource(label_main: Label):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())
    variant_list = []
    for variant_name in label_main.variants:
        found_variant = db.get_label_variant(name=label_main.name, variant=variant_name)
        variant_list.append(build_label_variant_resource(found_variant))

    return LabelResource(
        name=label_main.name,
        defaultVariant=label_main.default_variant,
        type="Label",
        allVariants=label_main.variants,
        variants=variant_list_to_dict(variant_list),
    ).to_dictionary()


def build_label_variant_resource(variant_data: LabelVariant):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())

    label_training_set_list = []
    training_set_variant_list = db.get_training_set_variant_from_label(
        label_name=variant_data.name, label_variant=variant_data.variant
    )
    for tsv in training_set_variant_list:
        label_training_set_list.append(build_training_set_variant_resource(tsv))

    label_variant_resource = LabelVariantResource(
        created=variant_data.created,
        description=variant_data.description,
        name=variant_data.name,
        owner=variant_data.owner,
        variant=variant_data.variant,
        entity=variant_data.entity,
        dataType=variant_data.value_type,
        status=variant_data.status,
        source={
            "Name": variant_data.source[0],
            "Variant": variant_data.source[1],
        },
        location={
            "entity": variant_data.location.entity,
            "value": variant_data.location.value,
            "timestamp": variant_data.location.timestamp,
        },
        trainingSets=resources_list_to_dict(label_training_set_list),
        tags=variant_data.tags if variant_data.tags is not None else [],
        properties=variant_data.properties
        if variant_data.properties is not None
        else [],
    ).to_dictionary()
    return label_variant_resource


def build_entity_resource(entity_main: Entity):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())

    entity_feature_list = []
    feature_list = db.get_features()
    for current_feature in feature_list:
        for variant_name in current_feature.variants:
            found_feature_variant = db.get_feature_variant(
                name=current_feature.name, variant=variant_name
            )
            if found_feature_variant.entity == entity_main.name:
                entity_feature_list.append(
                    build_feature_variant_resource(found_feature_variant)
                )

    entity_labels_list = []
    label_list = db.get_labels()
    for current_label in label_list:
        for variant_name in current_label.variants:
            found_label_variant = db.get_label_variant(
                name=current_label.name, variant=variant_name
            )
            if found_label_variant.entity == entity_main.name:
                entity_labels_list.append(
                    build_label_variant_resource(found_label_variant)
                )

    entity_training_set_list = []
    training_set_list = db.get_training_sets()
    for current_training_set in training_set_list:
        for variant_name in current_training_set.variants:
            found_training_set_variant = db.get_training_set_variant(
                name=current_training_set.name, variant=variant_name
            )
            entity_training_set_list.append(
                build_training_set_variant_resource(found_training_set_variant)
            )

    return EntityResource(
        name=entity_main.name,
        type=entity_main.type().capitalize(),
        description=entity_main.description,
        status=entity_main.status,
        features=resources_list_to_dict(entity_feature_list),
        labels=resources_list_to_dict(entity_labels_list),
        trainingSets=resources_list_to_dict(entity_training_set_list),
        tags=entity_main.tags if entity_main.tags is not None else [],
        properties=entity_main.properties if entity_main.properties is not None else [],
    ).to_dictionary()


def build_model_resource(model_obj: Model):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())

    model_feature_list = []
    feature_list = db.get_features()
    for current_feature in feature_list:
        for variant_name in current_feature.variants:
            found_variant = db.get_feature_variant(
                name=current_feature.name, variant=variant_name
            )
            if found_variant.name == model_obj.name:
                model_feature_list.append(build_feature_variant_resource(found_variant))

    model_label_list = []
    label_list = db.get_labels()
    for current_label in label_list:
        for variant_name in current_label.variants:
            found_variant = db.get_label_variant(
                name=current_label.name, variant=variant_name
            )
            if found_variant.variant == model_obj.name:
                model_label_list.append(build_label_variant_resource(found_variant))

    model_training_set_list = []
    training_set_list = db.get_training_sets()
    for current_training_set in training_set_list:
        for variant_name in current_training_set.variants:
            found_variant = db.get_training_set_variant(
                name=current_training_set.name, variant=variant_name
            )
            if found_variant.name == model_obj.name:
                model_training_set_list.append(
                    build_training_set_variant_resource(found_variant)
                )

    return ModelResource(
        name=model_obj.name,
        type="Model",
        description=model_obj.description,
        features=resources_list_to_dict(model_feature_list),
        labels=resources_list_to_dict(model_label_list),
        trainingSets=resources_list_to_dict(model_training_set_list),
        tags=model_obj.tags if model_obj.tags is not None else [],
        properties=model_obj.properties if model_obj.properties is not None else [],
    ).to_dictionary()


def build_user_resource(user_obj: User):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())

    user_feature_list = []
    feature_list = db.get_features()
    for current_feature in feature_list:
        for variant_name in current_feature.variants:
            found_variant = db.get_feature_variant(
                name=current_feature.name, variant=variant_name
            )
            if found_variant.owner == user_obj.name:
                user_feature_list.append(build_feature_variant_resource(found_variant))

    user_label_list = []
    label_list = db.get_labels()
    for current_label in label_list:
        for variant_name in current_label.variants:
            found_variant = db.get_label_variant(
                name=current_label.name, variant=variant_name
            )
            if found_variant.owner == user_obj.name:
                user_label_list.append(build_label_variant_resource(found_variant))

    user_training_set_list = []
    training_set_list = db.get_training_sets()
    for current_training_set in training_set_list:
        for variant_name in current_training_set.variants:
            found_variant = db.get_training_set_variant(
                name=current_training_set.name, variant=variant_name
            )
            if found_variant.owner == user_obj.name:
                user_training_set_list.append(
                    build_training_set_variant_resource(found_variant)
                )

    user_source_list = []
    source_list = db.get_sources()
    for current_source in source_list:
        for variant_name in current_source.variants:
            found_variant = db.get_source_variant(
                name=current_source.name, variant=variant_name
            )
            if found_variant.owner == user_obj.name:
                user_source_list.append(build_source_variant_resource(found_variant))

    return UserResource(
        name=user_obj.name,
        type="User",
        status=user_obj.status,
        features=resources_list_to_dict(user_feature_list),
        labels=resources_list_to_dict(user_label_list),
        trainingSets=resources_list_to_dict(user_training_set_list),
        sources=resources_list_to_dict(user_source_list),
        tags=user_obj.tags if user_obj.tags is not None else [],
        properties=user_obj.properties if user_obj.properties is not None else [],
    ).to_dictionary()


def build_provider_resource(provider_obj: Provider):
    db = MetadataRepositoryLocalImpl(SQLiteMetadata())
    provider_source_list = []
    source_list = db.get_sources()
    for current_source in source_list:
        for variant_name in current_source.variants:
            found_variant = db.get_source_variant(
                name=current_source.name, variant=variant_name
            )
            if found_variant.provider == provider_obj.name:
                provider_source_list.append(
                    build_source_variant_resource(found_variant)
                )

    provider_feature_list = []
    feature_list = db.get_features()
    for current_feature in feature_list:
        for variant_name in current_feature.variants:
            found_variant = db.get_feature_variant(
                name=current_feature.name, variant=variant_name
            )
            if found_variant.provider == provider_obj.name:
                provider_feature_list.append(
                    build_feature_variant_resource(found_variant)
                )

    provider_label_list = []
    label_list = db.get_labels()
    for current_label in label_list:
        for variant_name in current_label.variants:
            found_variant = db.get_label_variant(
                name=current_label.name, variant=variant_name
            )
            if found_variant.provider == provider_obj.name:
                provider_label_list.append(build_label_variant_resource(found_variant))

    return ProviderResource(
        name=provider_obj.name,
        type="Provider",
        description=provider_obj.description,
        providerType=provider_obj.config.type(),
        software=provider_obj.config.software(),
        team=provider_obj.team,
        sources=resources_list_to_dict(provider_source_list),
        status=provider_obj.status,
        features=resources_list_to_dict(provider_feature_list),
        labels=resources_list_to_dict(provider_label_list),
        tags=provider_obj.tags if provider_obj.tags is not None else [],
        properties=provider_obj.properties
        if provider_obj.properties is not None
        else [],
    ).to_dictionary()



import json
from .format import *
from .sqlite_metadata import *
from .resources import Model


def get_user_info_local(name):
    user = get_resource("user", name, True)

    format_rows("USER NAME: ", user["name"])
    format_rows("TYPE: ", user["type"])
    format_rows("STATUS: ", user["status"])
    tags = json.loads(user["tags"]) if user["tags"] is not None else []
    properties = (
        json.loads(user["properties"]) if user["properties"] is not None else {}
    )
    format_tags_and_properties(tags, properties)
    return {**user, "tags": tags, "properties": properties}


def get_entity_info_local(name):
    db = SQLiteMetadata()
    entity = get_resource("entity", name, True)

    returned_features_query_list = get_related_resources(
        "feature_variant", "entity", name
    )
    returned_features_list = format_resource_list(returned_features_query_list)

    returned_labels_query_list = get_related_resources("label_variant", "entity", name)
    returned_labels_list = format_resource_list(returned_labels_query_list)

    training_set_list = set()
    for f in returned_features_list:
        try:
            training_set_features_query_list = db.get_training_set_from_features(
                f["name"], f["variant"]
            )
        except ValueError:
            training_set_features_query_list = []
        for t in training_set_features_query_list:
            training_set_list.add(t)

    returned_training_sets_list = format_resource_list(
        training_set_list, "training_set_name", "training_set_variant"
    )

    returned_entity = {
        "name": name,
        "status": entity["status"],
        "features": returned_features_list,
        "labels": returned_labels_list,
        "trainingsets": returned_training_sets_list,
        "tags": json.loads(entity["tags"]) if entity["tags"] is not None else [],
        "properties": json.loads(entity["properties"])
        if entity["properties"] is not None
        else {},
    }

    format_rows(
        [
            ("ENTITY NAME: ", returned_entity["name"]),
            ("STATUS: ", returned_entity["status"]),
        ]
    )
    format_tags_and_properties(returned_entity["tags"], returned_entity["properties"])
    format_pg()
    format_rows("NAME", "VARIANT", "TYPE")
    for f in returned_entity["features"]:
        format_rows(f["name"], f["variant"], "feature")
    for l in returned_entity["labels"]:
        format_rows(l["name"], l["variant"], "label")
    for t in returned_entity["trainingsets"]:
        format_rows(t["name"], t["variant"], "training set")
    format_pg()
    return returned_entity


def get_resource_info_local(resource_type, name):
    resource = get_resource(resource_type, name)

    variants_list = get_variant_list(resource_type, name)

    returned_resource_list = {
        "name": resource["name"],
        "default_variant": resource["default_variant"],
        "variants": [(v["variant"], v["tags"], v["properties"]) for v in variants_list],
    }

    format_rows("NAME: ", returned_resource_list["name"])
    if "status" in resource:
        format_rows("STATUS: ", resource["status"])
    format_pg("VARIANTS:")
    formatted_variants = [("NAME", "DEFAULT", "TAGS", "PROPERTIES")]
    for v, t, p in returned_resource_list["variants"]:
        t = json.loads(t) if t is not None else []
        p = json.loads(p) if p is not None else {}
        default = "" if v != returned_resource_list["default_variant"] else "default"
        formatted_variants.append(
            (v, default, ", ".join(t), ", ".join([f"{k}:{v}" for (k, v) in p.items()]))
        )
    format_rows(formatted_variants)
    format_pg()
    return returned_resource_list


def get_feature_variant_info_local(name, variant):
    db = SQLiteMetadata()
    feature = get_variant("feature", name, variant)

    try:
        returned_training_sets_query_list = db.get_training_set_from_features(
            feature["name"], feature["variant"]
        )
    except ValueError:
        returned_training_sets_query_list = []
    returned_training_sets_list = format_resource_list(
        returned_training_sets_query_list, "training_set_name", "training_set_variant"
    )

    returned_feature = {
        "name": feature["name"],
        "variant": feature["variant"],
        "type": feature["data_type"],
        "entity": feature["entity"],
        "owner": feature["owner"],
        "description": feature["description"],
        "provider": feature["provider"],
        "status": feature["status"],
        "source": {
            "name": feature["source_name"],
            "variant": feature["source_variant"],
        },
        "trainingsets": returned_training_sets_list,
        "tags": json.loads(feature["tags"]) if feature["tags"] is not None else [],
        "properties": json.loads(feature["properties"])
        if feature["properties"] is not None
        else {},
    }
    format_rows(
        [
            ("NAME: ", returned_feature["name"]),
            ("VARIANT: ", returned_feature["variant"]),
            ("TYPE:", returned_feature["type"]),
            ("ENTITY:", returned_feature["entity"]),
            ("OWNER:", returned_feature["owner"]),
            ("DESCRIPTION:", returned_feature["description"]),
            ("PROVIDER:", returned_feature["provider"]),
            ("STATUS: ", returned_feature["status"]),
        ]
    )
    format_tags_and_properties(returned_feature["tags"], returned_feature["properties"])
    format_pg("SOURCE: ")
    format_rows(
        [
            ("NAME", "VARIANT"),
            (returned_feature["source"]["name"], returned_feature["source"]["variant"]),
        ]
    )
    format_pg("TRAINING SETS:")
    format_rows("NAME", "VARIANT")
    for t in returned_feature["trainingsets"]:
        format_rows(t["name"], t["variant"])
    format_pg()
    return returned_feature


def get_label_variant_info_local(name, variant):
    db = SQLiteMetadata()
    label = get_variant("label", name, variant)

    try:
        returned_training_sets_query_list = db.get_training_set_from_labels(
            label["name"], label["variant"]
        )
    except ValueError:
        returned_training_sets_query_list = []

    returned_training_sets_list = format_resource_list(
        returned_training_sets_query_list
    )

    returned_label = {
        "name": label["name"],
        "variant": label["variant"],
        "type": label["data_type"],
        "entity": label["entity"],
        "owner": label["owner"],
        "description": label["description"],
        "provider": label["provider"],
        "status": label["status"],
        "source": {"name": label["source_name"], "variant": label["source_variant"]},
        "trainingsets": returned_training_sets_list,
        "tags": json.loads(label["tags"]) if label["tags"] is not None else [],
        "properties": json.loads(label["properties"])
        if label["properties"] is not None
        else {},
    }
    format_rows(
        [
            ("NAME: ", returned_label["name"]),
            ("VARIANT: ", returned_label["variant"]),
            ("TYPE:", returned_label["type"]),
            ("ENTITY:", returned_label["entity"]),
            ("OWNER:", returned_label["owner"]),
            ("DESCRIPTION:", returned_label["description"]),
            ("PROVIDER:", returned_label["provider"]),
            ("STATUS: ", returned_label["status"]),
        ]
    )
    format_tags_and_properties(returned_label["tags"], returned_label["properties"])
    format_pg("SOURCE: ")
    format_rows(
        [
            ("NAME", "VARIANT"),
            (returned_label["source"]["name"], returned_label["source"]["variant"]),
        ]
    )
    format_pg("TRAINING SETS:")
    format_rows("NAME", "VARIANT")
    for t in returned_label["trainingsets"]:
        format_rows(t["name"], t["variant"])
    format_pg()
    return returned_label


def get_source_variant_info_local(name, variant):
    db = SQLiteMetadata()
    source = get_variant("source", name, variant)

    try:
        returned_features_query_list = db.get_feature_variants_from_source(
            name, variant
        )
        returned_features_list = format_resource_list(returned_features_query_list)
    except ValueError:
        returned_features_list = []

    try:
        returned_labels_query_list = db.get_label_variants_from_source(name, variant)
        returned_labels_list = format_resource_list(returned_labels_query_list)
    except ValueError:
        returned_labels_list = []

    training_set_list = set()

    for f in returned_features_list:
        try:
            training_set_features_query_list = db.get_training_set_from_features(
                f["name"], f["variant"]
            )
        except ValueError:
            training_set_features_query_list = []
        for t in training_set_features_query_list:
            training_set_list.add(t)

    for l in returned_labels_list:
        try:
            training_set_labels_query_list = db.get_training_set_variant_from_label(
                l["name"], l["variant"]
            )
        except ValueError:
            training_set_labels_query_list = []
        for t in training_set_labels_query_list:
            training_set_list.add(t)

    returned_training_sets_list = format_resource_list(training_set_list)

    returned_source = {
        "name": source["name"],
        "variant": source["variant"],
        "owner": source["owner"],
        "description": source["description"],
        "provider": source["provider"],
        "status": source["status"],
        "definition": source["definition"],
        "features": returned_features_list,
        "labels": returned_labels_list,
        "trainingsets": returned_training_sets_list,
        "tags": json.loads(source["tags"]) if source["tags"] is not None else [],
        "properties": json.loads(source["properties"])
        if source["properties"] is not None
        else {},
    }

    format_rows(
        [
            ("NAME: ", returned_source["name"]),
            ("VARIANT: ", returned_source["variant"]),
            ("OWNER:", returned_source["owner"]),
            ("DESCRIPTION:", returned_source["description"]),
            ("PROVIDER:", returned_source["provider"]),
            ("STATUS: ", returned_source["status"]),
        ]
    )
    format_tags_and_properties(returned_source["tags"], returned_source["properties"])
    format_pg("DEFINITION:")
    print(returned_source["definition"])
    print("FEATURES:")
    format_rows("NAME", "VARIANT")
    for f in returned_source["features"]:
        format_rows(f["name"], f["variant"])
    format_pg("LABELS:")
    format_rows("NAME", "VARIANT")
    for l in returned_source["labels"]:
        format_rows(l["name"], l["variant"])
    format_pg("TRAINING SETS:")
    format_rows("NAME", "VARIANT")
    for t in returned_source["trainingsets"]:
        format_rows(t["name"], t["variant"])
    format_pg()
    return returned_source


def get_training_set_variant_info_local(name, variant):
    db = SQLiteMetadata()
    training_set = get_variant("training-set", name, variant)

    try:
        returned_features_query_list = db.get_training_set_features(name, variant)
    except ValueError:
        returned_features_query_list = []
    returned_features_list = format_resource_list(
        returned_features_query_list, "feature_name", "feature_variant"
    )

    returned_training_set = {
        "name": training_set["name"],
        "variant": training_set["variant"],
        "owner": training_set["owner"],
        "description": training_set["description"],
        "status": training_set["status"],
        "label": {
            "name": training_set["label_name"],
            "variant": training_set["label_variant"],
        },
        "features": returned_features_list,
        "tags": json.loads(training_set["tags"])
        if training_set["tags"] is not None
        else [],
        "properties": json.loads(training_set["properties"])
        if training_set["properties"] is not None
        else {},
    }

    format_rows(
        [
            ("NAME: ", returned_training_set["name"]),
            ("VARIANT: ", returned_training_set["variant"]),
            ("OWNER:", returned_training_set["owner"]),
            ("DESCRIPTION:", returned_training_set["description"]),
            ("STATUS: ", returned_training_set["status"]),
        ]
    )
    format_tags_and_properties(
        returned_training_set["tags"], returned_training_set["properties"]
    )
    format_pg("LABEL: ")
    format_rows(
        [
            ("NAME", "VARIANT"),
            (
                returned_training_set["label"]["name"],
                returned_training_set["label"]["variant"],
            ),
        ]
    )
    format_pg("FEATURES:")
    format_rows("NAME", "VARIANT")
    for f in returned_training_set["features"]:
        format_rows(f["name"], f["variant"])
    format_pg()
    return returned_training_set


def get_provider_info_local(name):
    db = SQLiteMetadata()
    provider = get_resource("provider", name, True)

    try:
        returned_features_query_list = db.get_feature_variants_from_provider(name)
        returned_features_list = format_resource_list(returned_features_query_list)
    except ValueError:
        returned_features_list = []

    try:
        returned_labels_query_list = db.get_label_variants_from_provider(name)
        returned_labels_list = format_resource_list(returned_labels_query_list)
    except ValueError:
        returned_labels_list = []

    returned_provider = {
        "name": provider["name"],
        "type": provider["type"],
        "description": provider["description"],
        "type": provider["provider_type"],
        "software": provider["software"],
        "team": provider["team"],
        "status": provider["status"],
        "serializedConfig": provider["serialized_config"],
        "sources": provider["sources"],
        "features": returned_features_list,
        "labels": returned_labels_list,
        "tags": json.loads(provider["tags"]) if provider["tags"] is not None else [],
        "properties": json.loads(provider["properties"])
        if provider["properties"] is not None
        else {},
    }

    format_rows(
        [
            ("NAME: ", returned_provider["name"]),
            ("DESCRIPTION: ", returned_provider["description"]),
            ("TYPE: ", returned_provider["type"]),
            ("SOFTWARE: ", returned_provider["software"]),
            ("TEAM: ", returned_provider["team"]),
            ("STATUS: ", returned_provider["status"]),
        ]
    )
    format_tags_and_properties(
        returned_provider["tags"], returned_provider["properties"]
    )
    format_rows("SOURCE:", returned_provider["sources"])
    format_pg("FEATURES:")
    format_rows("NAME", "VARIANT")
    for f in returned_provider["features"]:
        format_rows(f["name"], f["variant"])
    format_pg("LABELS:")
    format_rows("NAME", "VARIANT")
    for l in returned_provider["labels"]:
        format_rows(l["name"], l["variant"])
    format_pg()
    return returned_provider


def get_related_resources(table, column, name):
    db = SQLiteMetadata()
    try:
        res_list = db.query_resource(table, column, name)
    except ValueError:
        res_list = []
    return res_list


def get_resource(resource_type, name, should_fetch_tags_properties=False):
    db = SQLiteMetadata()
    resource_sql_functions = {
        "user": db.get_user,
        "entity": db.get_entity,
        "feature": db.get_feature,
        "label": db.get_label,
        "source": db.get_source,
        "training-set": db.get_training_set,
        "model": db.get_model,
        "provider": db.get_provider,
    }
    res_list = resource_sql_functions[resource_type](name, should_fetch_tags_properties)
    return res_list


def get_variant_list(resource_type, name):
    db = SQLiteMetadata()
    variant_list_sql_functions = {
        "feature": db.get_feature_variants_from_feature,
        "label": db.get_label_variants_from_label,
        "source": db.get_source_variants_from_source,
        "training-set": db.get_training_set_variant_from_training_set,
    }

    res_list = variant_list_sql_functions[resource_type](name)
    return res_list


def get_variant(resource_type, name, variant):
    db = SQLiteMetadata()

    variant_sql_functions = {
        "feature": db.get_feature_variant,
        "label": db.get_label_variant,
        "source": db.get_source_variant,
        "training-set": db.get_training_set_variant,
    }

    res_list = variant_sql_functions[resource_type](name, variant)
    return res_list


def format_resource_list(res_list, name_col="name", var_col="variant"):
    formatted_res = [{"name": r[name_col], "variant": r[var_col]} for r in res_list]
    return formatted_res


def get_model_info_local(name) -> Model:
    model = get_resource("model", name)

    format_rows("MODEL NAME: ", model["name"])
    format_rows("TYPE: ", model["type"])

    return Model(model["name"], description="", tags=[], properties={})


def format_tags_and_properties(tags, properties):
    if len(tags):
        format_rows("TAGS:", ", ".join(tags))
    if len(properties):
        format_rows(
            "PROPERTIES:", ", ".join([f"{k}:{v}" for (k, v) in properties.items()])
        )
